from enum import unique
from ..bases.xmi_base_enum import XmiBaseEnum

@unique
class XmiStructuralSurfaceMemberTypeEnum(XmiBaseEnum):
    SLAB = "Slab"
    WALL = "Wall"
    PAD_FOOTING = "Pad Footing"
    STRIP_FOOTING = "Strip Footing"
    PILECAP = "Pilecap"  # i think this shouldn't be here
    ROOF_PANEL = "Roof Panel"  # this sounds like archi. should not be here
    WALL_PANEL = "Wall Panel"  # this sounds like archi. should not be here
    # For future changes to be shifted to foundation type elements
    RAFT = "Raft"


# Testing run python -m src.xmi.v2.models.enums.xmi_structural_surface_member_type_enum

print(XmiStructuralSurfaceMemberTypeEnum.STRIP_FOOTING.value)
print(XmiStructuralSurfaceMemberTypeEnum.SLAB.name)


from enum import unique
from ..bases.xmi_base_enum import XmiBaseEnum

@unique
class XmiStructuralSurfaceMemberSpanTypeEnum(XmiBaseEnum):
    ONE_WAY = "One Way"
    TWO_WAY = "Two Way"


# Testing run python -m src.xmi.v2.models.enums.xmi_structural_surface_member_span_type_enum

print(XmiStructuralSurfaceMemberSpanTypeEnum.ONE_WAY.value)
print(XmiStructuralSurfaceMemberSpanTypeEnum.TWO_WAY.name)


# Optional, for forward declarations in Python 3.7+
from __future__ import annotations

from ..entities.xmi_segment import XmiSegment
from ..entities.xmi_structural_cross_section import XmiStructuralCrossSection

from .xmi_structural_point_connection import XmiStructuralPointConnection
from ..xmi_base import XmiBaseEntity
from ..enums.xmi_structural_curve_member_enums import *
from ..geometries.xmi_base_geometry import XmiBaseGeometry

from ..xmi_errors import *


class XmiStructuralCurveMember(XmiBaseEntity):
    __slots__ = XmiBaseEntity.__slots__ + \
        ('_cross_section',
         '_storey',
         '_curve_member_type',
         '_nodes',
         '_segments',
         '_system_line',
         '_begin_node',
         '_end_node',
         '_length',
         '_local_axis_x',
         '_local_axis_y',
         '_local_axis_z',
         '_begin_node_x_offset',
         '_end_node_x_offset',
         '_begin_node_y_offset',
         '_end_node_y_offset',
         '_begin_node_z_offset',
         '_end_node_z_offset',
         '_end_fixity_start',
         '_end_fixity_end',
         )

    _attributes_needed = [slot[1:] if slot.startswith(
        '_') else slot for slot in __slots__ if slot != "_entity_type"]

    def __init__(self,
                 cross_section: XmiStructuralCrossSection,
                 curve_member_type: XmiStructuralCurveMemberTypeEnum,
                 system_line: XmiStructuralCurveMemberSystemLineEnum,
                 local_axis_x: tuple = (1.0, 0.0, 0.0),
                 local_axis_y: tuple = (0.0, 1.0, 0.0),
                 local_axis_z: tuple = (0.0, 0.0, 1.0),
                 begin_node_x_offset: float | int = 0.0,
                 end_node_x_offset: float | int = 0.0,
                 begin_node_y_offset: float | int = 0.0,
                 end_node_y_offset: float | int = 0.0,
                 begin_node_z_offset: float | int = 0.0,
                 end_node_z_offset: float | int = 0.0,
                 segments: list[XmiBaseGeometry] = [],
                 nodes: list[XmiStructuralPointConnection] = [],
                 length: int | float | None = None,
                 begin_node: XmiStructuralPointConnection = None,
                 end_node: XmiStructuralPointConnection = None,
                 #  end_fixity_start,  # optional
                 #  end_fixity_end,  # optional
                 storey: str | None = None,
                 id: str = None,
                 name: str = None,
                 description: str = None,
                 ifcguid: str = None,
                 **kwargs
                 ):
        entity_type = "XmiStructuralCurveMember"

        # Check for mutual exclusivity, things that are optional should be inside any
        # if kwargs and any([
        #         length,
        #         storey,
        #         id, name, description, ifcguid
        # ]):
        #     raise ValueError(
        #         "Please use either standard parameters or kwargs, not both.")

        # Ensure cross_section is provided
        if cross_section is None:
            raise ValueError(
                "The 'cross_section' parameter is compulsory and must be provided.")

        # Ensure curve_member_type is provided
        if curve_member_type is None:
            raise ValueError(
                "The 'curve_member_type' parameter is compulsory and must be provided.")

        # Ensure system_line is provided
        if system_line is None:
            raise ValueError(
                "The 'system_line' parameter is compulsory and must be provided.")

        # Ensure nodes is provided
        if nodes is None:
            raise ValueError(
                "The 'nodes' parameter is compulsory and must be provided.")

        # Ensure segments is provided
        if segments is None:
            # if segments is None and not in kwargs, then check if the nodes and segment_types is provided
            # if not provided then only raise error
            raise ValueError(
                "The 'segments' parameter is compulsory and must be provided.")

        # Ensure begin_node is provided
        if begin_node is None:
            raise ValueError(
                "The 'begin_node' parameter is compulsory and must be provided.")

        # Ensure end_node is provided
        if end_node is None:
            raise ValueError(
                "The 'end_node' parameter is compulsory and must be provided.")

        # Initialize parent class
        super().__init__(id=id,
                         name=name,
                         ifcguid=ifcguid,
                         description=description,
                         entity_type=entity_type
                         )

        # Initialize attributes
        self.set_attributes(
            cross_section=cross_section,
            curve_member_type=curve_member_type,
            nodes=nodes,
            segments=segments,
            system_line=system_line,
            begin_node=begin_node,
            end_node=end_node,
            local_axis_x=local_axis_x,
            local_axis_y=local_axis_y,
            local_axis_z=local_axis_z,
            begin_node_x_offset=begin_node_x_offset,
            end_node_x_offset=end_node_x_offset,
            begin_node_y_offset=begin_node_y_offset,
            end_node_y_offset=end_node_y_offset,
            begin_node_z_offset=begin_node_z_offset,
            end_node_z_offset=end_node_z_offset,
            length=length,
            storey=storey,
            # segment_types=segment_types,
            ** kwargs)

    def set_attributes(self,
                       cross_section: XmiStructuralCrossSection,
                       curve_member_type: XmiStructuralCurveMemberTypeEnum,
                       system_line: XmiStructuralCurveMemberSystemLineEnum,
                       nodes: list[XmiStructuralPointConnection],
                       segments: list[XmiSegment],
                       #    segment_types: list[XmiSegmentTypeEnum],
                       begin_node: XmiStructuralPointConnection,
                       end_node: XmiStructuralPointConnection,
                       local_axis_x,
                       local_axis_y,
                       local_axis_z,
                       begin_node_x_offset,
                       end_node_x_offset,
                       begin_node_y_offset,
                       end_node_y_offset,
                       begin_node_z_offset,
                       end_node_z_offset,
                       length,
                       storey,
                       **kwargs):

        attributes = [
            ('cross_section', cross_section),
            ('curve_member_type', curve_member_type),
            ('nodes', nodes),
            ('segments', segments),
            ('system_line', system_line),
            ('begin_node', begin_node),
            ('end_node', end_node),
            ('local_axis_x', local_axis_x),
            ('local_axis_y', local_axis_y),
            ('local_axis_z', local_axis_z),
            ('begin_node_x_offset', begin_node_x_offset),
            ('end_node_x_offset', end_node_x_offset),
            ('begin_node_y_offset', begin_node_y_offset),
            ('end_node_y_offset', end_node_y_offset),
            ('begin_node_z_offset', begin_node_z_offset),
            ('end_node_z_offset', end_node_z_offset),
            ('length', length),
            ('storey', storey)
        ]

        for attr_name, attr_value in attributes:
            value = kwargs.get(attr_name, attr_value)
            try:
                setattr(self, attr_name, value)
            except AttributeError as e:
                print(
                    f"Caught an AttributeError while setting {attr_name}: {e}")
                setattr(self, attr_name, None)

    @property
    def storey(self):
        return self._storey

    @storey.setter
    def storey(self, value):
        if not isinstance(value, str):
            raise TypeError("storey attribute should be an str")
        self._storey = value

    @property
    def curve_member_type(self):
        return self._curve_member_type

    @curve_member_type.setter
    def curve_member_type(self, value):
        if not isinstance(value, (XmiStructuralCurveMemberTypeEnum)):
            raise TypeError(
                "curve_member_type should be type XmiStructuralCurveMemberTypeEnum")
        self._curve_member_type = value

    @property
    def nodes(self):
        return self._nodes

    @nodes.setter
    def nodes(self, value):
        if not isinstance(value, list):
            raise TypeError("nodes attribute should be a list")
        for item in value:
            if not isinstance(item, XmiStructuralPointConnection):
                raise ValueError(
                    f"All items must be instances of XmiStructuralPointConnection, got {type(item)} instead.")
        self._nodes = value

    @property
    def begin_node(self):
        return self._begin_node

    @begin_node.setter
    def begin_node(self, value):
        if not isinstance(value, XmiStructuralPointConnection):
            raise TypeError(
                "begin_node should be of type XmiStructuralPointConnection")
        self._begin_node = value

    @property
    def end_node(self):
        return self._end_node

    @end_node.setter
    def end_node(self, value):
        if not isinstance(value, XmiStructuralPointConnection):
            raise TypeError(
                "EndNode should be of type XmiStructuralPointConnection")
        self._end_node = value

    @property
    def cross_section(self):
        return self._cross_section

    @cross_section.setter
    def cross_section(self, value):
        if not isinstance(value, XmiStructuralCrossSection):
            raise TypeError(
                "cross_section should be of type XmiStructuralCrossSection")
        self._cross_section = value

    @property
    def system_line(self):
        return self._system_line

    @system_line.setter
    def system_line(self, value):
        if not isinstance(value, XmiStructuralCurveMemberSystemLineEnum):
            raise TypeError(
                "system_line should be of type XmiStructuralCurveMemberSystemLineEnum")
        self._system_line = value

    @property
    def segments(self):
        return self._segments

    @segments.setter
    def segments(self, value):
        if not isinstance(value, list):
            raise TypeError(
                "Segments should be of type list")

        # if len(value) < 1:
        #     raise ValueError(
        #         "'segments' list should have at least 1 items of type XmiSegment")

        for item in value:
            if not isinstance(item, XmiSegment):
                raise ValueError(
                    f"All items must be instances of XmiSegment, got {type(item)} instead.")
        self._segments = value

    @property
    def length(self):
        return self._length

    @length.setter
    def length(self, value):
        if not isinstance(value, (int, float, None)):
            raise TypeError(
                "length should be of type float or integer")
        self._length = value

    @property
    def local_axis_x(self):
        return self._local_axis_x

    @local_axis_x.setter
    def local_axis_x(self, value):
        if not isinstance(value, tuple):
            raise TypeError(
                "local_axis_x should be type XmiStructuralCurveMemberTypeEnum")
        self._local_axis_x = value

    @property
    def local_axis_y(self):
        return self._local_axis_y

    @local_axis_y.setter
    def local_axis_y(self, value):
        if not isinstance(value, tuple):
            raise TypeError(
                "local_axis_y should be type XmiStructuralCurveMemberTypeEnum")
        self._local_axis_y = value

    @property
    def local_axis_z(self):
        return self._local_axis_z

    @local_axis_z.setter
    def local_axis_z(self, value):
        if not isinstance(value, tuple):
            raise TypeError(
                "local_axis_z should be type XmiStructuralCurveMemberTypeEnum")
        self._local_axis_z = value

    @property
    def begin_node_x_offset(self):
        return self._begin_node_x_offset

    @begin_node_x_offset.setter
    def begin_node_x_offset(self, value):
        if not isinstance(value, (float, int)):
            raise TypeError(
                "begin_node_x_offset should be of type float or int")
        self._begin_node_x_offset = value

    @property
    def begin_node_y_offset(self):
        return self._begin_node_y_offset

    @begin_node_y_offset.setter
    def begin_node_y_offset(self, value):
        if not isinstance(value, (float, int)):
            raise TypeError(
                "begin_node_y_offset should be of type float or int")
        self._begin_node_y_offset = value

    @property
    def begin_node_z_offset(self):
        return self._begin_node_z_offset

    @begin_node_z_offset.setter
    def begin_node_z_offset(self, value):
        if not isinstance(value, (float, int)):
            raise TypeError(
                "begin_node_z_offset should be of type float or int")
        self._begin_node_z_offset = value

    @property
    def end_node_x_offset(self):
        return self._end_node_x_offset

    @end_node_x_offset.setter
    def end_node_x_offset(self, value):
        if not isinstance(value, (float, int)):
            raise TypeError(
                "end_node_x_offset should be of type float or int")
        self._end_node_x_offset = value

    @property
    def end_node_y_offset(self):
        return self._end_node_y_offset

    @end_node_y_offset.setter
    def end_node_y_offset(self, value):
        if not isinstance(value, (float, int)):
            raise TypeError(
                "end_node_y_offset should be of type float or int")
        self._end_node_y_offset = value

    @property
    def end_node_z_offset(self):
        return self._end_node_z_offset

    @end_node_z_offset.setter
    def end_node_z_offset(self, value):
        if not isinstance(value, (float, int)):
            raise TypeError(
                "end_node_z_offset should be of type float or int")
        self._end_node_z_offset = value

    @property
    def end_fixity_start(self):
        return self._end_fixity_start

    @end_fixity_start.setter
    def end_fixity_start(self, value):
        if not isinstance(value, (float, int)):
            raise TypeError(
                "end_fixity_start should be of type float or int")
        self._end_fixity_start = value

    @property
    def end_fixity_end(self):
        return self._end_fixity_end

    @end_fixity_end.setter
    def end_fixity_end(self, value):
        if not isinstance(value, (float, int)):
            raise TypeError(
                "end_fixity_end should be of type float or int")
        self._end_fixity_end = value

    def is_empty_or_whitespace(input_string: str) -> bool:
        return not input_string or not input_string.strip()

    @classmethod
    def convert_local_axis_string_to_tuple(cls, axis_direction, local_axis_str: str) -> tuple:
        local_axis_list: list[str] = local_axis_str.split(',')
        if len(local_axis_list) != 3:
            raise XmiMissingRequiredAttributeError(
                f"The XmiStructuralCurveMember 'local_axis_{axis_direction}' attribute should have 3 parameters")

        for local_axis_value in local_axis_list:
            if cls.is_empty_or_whitespace(local_axis_value):
                raise XmiInconsistentDataTypeError(
                    f"The individual parameter [{local_axis_value}] within the XmiStructuralCurveMember 'local_axis_{axis_direction}' attribute should not be empty string or empty space")
            try:
                float_value = float(local_axis_value)
            except ValueError:
                raise XmiInconsistentDataTypeError(
                    f"The parameter [{local_axis_value}] within the XmiStructuralCurveMember 'local_axis_{axis_direction}' attribute should be convertible to float")

        parameter_tuple = tuple([float(param) for param in local_axis_list])
        return parameter_tuple

    @classmethod
    def from_dict(cls, obj: dict) -> XmiStructuralCurveMember:
        instance = None
        exceptions = []
        processed_data = obj.copy()

        for attr in cls._attributes_needed:
            if attr not in processed_data:
                exceptions.append(Exception(f"Missing attribute: {attr}"))
                processed_data[attr] = None

        # for type conversion when reading dictionary
        try:
            # check for cross_section_found
            cross_section_found = processed_data['cross_section']
            if cross_section_found is None:
                exceptions.append(XmiMissingReferenceInstanceError(
                    "Please provide cross_section value of type XmiStructuralCrossSection"))
                return None, exceptions
            if not isinstance(cross_section_found, XmiStructuralCrossSection):
                exceptions.append(XmiInconsistentDataTypeError(
                    "cross_section provided need to be of instance XmiStructuralCrossSection"))
                return None, exceptions

            # check for nodes
            nodes_found = processed_data['nodes']
            if nodes_found is None:
                exceptions.append(XmiMissingRequiredAttributeError(
                    "Please provide value for the nodes attribute"))
                return None, exceptions

            if not isinstance(nodes_found, list):
                exceptions.append(XmiInconsistentDataTypeError(
                    "nodes value provided need to be of instance list"))
                return None, exceptions
            for node in nodes_found:
                if not isinstance(node, XmiStructuralPointConnection):
                    exceptions.append(XmiInconsistentDataTypeError(
                        "nodes value provided need to be of instance XmiStructuralPointConnection"))
                    return None, exceptions

            # check for segments
            segments_found = processed_data['segments']
            if segments_found is None:
                exceptions.append(XmiMissingRequiredAttributeError(
                    "Please provide value for the segments attribute"))
                return None, exceptions

            # if segment is not type list, return error
            if not isinstance(segments_found, list):
                exceptions.append(XmiInconsistentDataTypeError(
                    "segments value provided need to be of instance list"))
                return None, exceptions

            # if segments length = 0, return error
            if len(segments_found) == 0:
                exceptions.append(XmiMissingRequiredAttributeError(
                    "The 'segments' parameter requires at least 1 segment of Type XmiSegment"))
                return None, exceptions

            # check for all segment datatypes
            for segment in segments_found:
                if not isinstance(segment, XmiSegment):
                    exceptions.append(XmiInconsistentDataTypeError(
                        "segment value provided need to be of instance XmiSegment"))
                    return None, exceptions

            # setting up begin_node and end_node of element
            if len(segments_found) > 0:
                segment_begin: XmiSegment = segments_found[0]
                segment_end: XmiSegment = segments_found[len(
                    segments_found) - 1]

                begin_node_found = next(
                    (spc for spc in nodes_found if spc == segment_begin.begin_node), None)
                end_node_found = next(
                    (spc for spc in nodes_found if spc == segment_end.end_node), None)

                # check against xmi_dict provided
                if begin_node_found.name != processed_data['begin_node']:
                    exceptions.append(ValueError(
                        "'begin_node' in dictionary differs than the 1 defined in dictionary"))
                    return None, exceptions
                if end_node_found.name != processed_data['end_node']:
                    exceptions.append(ValueError(
                        "'end_node' in dictionary differs than the 1 defined in dictionary"))
                    return None, exceptions

                processed_data['begin_node'] = begin_node_found
                processed_data['end_node'] = end_node_found

            # check for local_axis_x
            local_axis_x_found = processed_data['local_axis_x']
            if local_axis_x_found is None:
                exceptions.append(XmiMissingRequiredAttributeError(
                    "Please provide value for the parameters attribute"))
                return None, exceptions

            local_axis_x_found = XmiStructuralCurveMember.convert_local_axis_string_to_tuple(
                'x', processed_data['local_axis_x'])

            if not isinstance(local_axis_x_found, tuple):
                exceptions.append(XmiInconsistentDataTypeError(
                    "local_axis_x value after conversion using the convert_parameter_string_to_tuple function should be of type tuple"))
                return None, exceptions
            processed_data['local_axis_x'] = local_axis_x_found

            # check for local_axis_y
            local_axis_y_found = processed_data['local_axis_y']
            if local_axis_y_found is None:
                exceptions.append(XmiMissingRequiredAttributeError(
                    "Please provide value for the parameters attribute"))
                return None, exceptions
            local_axis_y_found = XmiStructuralCurveMember.convert_local_axis_string_to_tuple(
                'y', processed_data['local_axis_y'])

            if not isinstance(local_axis_y_found, tuple):
                exceptions.append(XmiInconsistentDataTypeError(
                    "local_axis_y value after conversion using the convert_parameter_string_to_tuple function should be of type tuple"))
                return None, exceptions
            processed_data["local_axis_y"] = local_axis_y_found

            # check for local_axis_z
            local_axis_z_found = processed_data['local_axis_z']
            if local_axis_z_found is None:
                exceptions.append(XmiMissingRequiredAttributeError(
                    "Please provide value for the parameters attribute"))
                return None, exceptions
            local_axis_z_found = XmiStructuralCurveMember.convert_local_axis_string_to_tuple(
                'z', processed_data['local_axis_z'])
            if not isinstance(local_axis_z_found, tuple):
                exceptions.append(XmiInconsistentDataTypeError(
                    "local_axis_z value after conversion using the convert_parameter_string_to_tuple function should be of type tuple"))
                return None, exceptions
            processed_data["local_axis_z"] = local_axis_z_found

            # check system_line
            system_line_found = processed_data['system_line']
            if system_line_found is None:
                exceptions.append(XmiMissingRequiredAttributeError(
                    "Please provide value for the system_line attribute"))
                return None, exceptions
            system_line_found = XmiStructuralCurveMemberSystemLineEnum.from_attribute_get_enum(
                processed_data['system_line'])
            processed_data["system_line"] = system_line_found

            # check curve_member_type
            curve_member_type_found = processed_data['curve_member_type']
            if curve_member_type_found is None:
                exceptions.append(XmiMissingRequiredAttributeError(
                    "Please provide value for the curve_member_type attribute"))
                return None, exceptions
            curve_member_type_found = XmiStructuralCurveMemberTypeEnum.from_attribute_get_enum(
                processed_data['curve_member_type'])
            processed_data["curve_member_type"] = curve_member_type_found

        except KeyError as e:
            exceptions.append(e)
            return None, exceptions

        try:
            instance = cls(
                ** processed_data)
        except Exception as e:
            exceptions.append(
                Exception(f"Error instantiating XmiStructuralCurveMember: {obj}"))

        return instance, exceptions

    # additional parameters are used to inject reference elements
    @classmethod
    def from_xmi_dict_obj(cls, xmi_dict_obj: dict,
                          cross_section: XmiStructuralCrossSection = None,
                          nodes: list[XmiStructuralPointConnection] = None,
                          segments: list[XmiBaseEntity] = None,
                          ) -> XmiStructuralCurveMember:
        # Define a mapping from snake_case keys to custom keys
        KEY_MAPPING = {
            "CrossSection": "cross_section",
            "Storey": "storey",
            "Type": "curve_member_type",
            "Nodes": "nodes",
            "Segments": "segments",
            "SystemLine": "system_line",
            "BeginNode": "begin_node",
            "EndNode": "end_node",
            "Length": "length",
            "LocalAxisX": "local_axis_x",
            "LocalAxisY": "local_axis_y",
            "LocalAxisZ": "local_axis_z",
            "BeginNodeXOffset": "begin_node_x_offset",
            "EndNodeXOffset": "end_node_x_offset",
            "BeginNodeYOffset": "begin_node_y_offset",
            "EndNodeYOffset": "end_node_y_offset",
            "BeginNodeZOffset": "begin_node_z_offset",
            "EndNodeZOffset": "end_node_z_offset",
            "EndFixityStart": "end_fixity_start",
            "EndFixityEnd": "end_fixity_end",
            "Name": "name",
            "Description": "description",
            "ID": "id",
            "IFCGUID": "ifcguid",
            # "CircularArcCentre": "circular_arc_centre", # NOT IN USE
            # "CircularArcRadius": "circular_arc_radius", # NOT IN USE
            # "StiffnessModifierArea": "stiffness_modifier_area",
            # "StiffnessModifierAsy": "stiffness_modifier_shear_local_y_axis",
            # "StiffnessModifierAsz": "stiffness_modifier_shear_local_z_axis",
            # "StiffnessModifierTorsion": "stiffness_modifier_shear_torsion",
            # "StiffnessModifierIyy": "stiffness_modifier_moment_local_y_axis",
            # "StiffnessModifierIzz": "stiffness_modifier_moment_local_z_axis",
            # "StiffnessModifierMass": "stiffness_modifier_moment_mass",
            # "StiffnessModifierWeight": "stiffness_modifier_moment_weight",
            # "EndFixityAxialStart": "end_fixity_axial_start",
            # "EndFixityShearMajorStart": "end_fixity_shear_major_start",
            # "EndFixityShearMinorStart": "end_fixity_shear_minor_start",
            # "EndFixityTorsionStart": "end_fixity_torsion_start",
            # "EndFixityMomentMajorStart": "end_fixity_moment_major_start",
            # "EndFixityMomentMinorStart": "end_fixity_moment_minor_start",
            # "EndFixityAxialEnd": "end_fixity_axial_end",
            # "EndFixityShearMajorEnd": "end_fixity_shear_major_end",
            # "EndFixityShearMinorEnd": "end_fixity_shear_minor_end",
            # "EndFixityTorsionEnd": "end_fixity_torsion_end/",
            # "EndFixityMomentMajorEnd": "begin_node_y_offset",
            # "EndFixityMomentMinorEnd": "end_node_y_offset",
            # "LateralRestrain": "begin_node",
            # "LateralRestrainLocation": "end_node",
            # "LengthEffectiveMajor": "length",
            # "LengthEffectiveMinor": "local_axis_x",
            # "SwayInMajor": "local_axis_y",
            # "SwayInMinor": "local_axis_z",
            # "BracedAbtMajor": "begin_node_y_offset",
            # "BracedAbtMinor": "end_node_y_offset",
        }

        instance: XmiStructuralCurveMember | None = None
        exceptions: list[Exception] = []
        processed_data: dict = {KEY_MAPPING.get(
            key, key): value for key, value in xmi_dict_obj.items()}

        if cross_section is not None:
            processed_data['cross_section'] = cross_section

        if nodes is not None:
            processed_data['nodes'] = nodes

        if segments is not None:
            processed_data['segments'] = segments

        # if segment_types is not None:
        #     processed_data['segment_types'] = segment_types

        instance, exceptions_found = cls.from_dict(
            processed_data)

        exceptions.extend(exceptions_found)

        return instance, exceptions



from typing import Optional, Tuple, List
from pydantic import Field, field_validator, model_validator
from ..bases.xmi_base_geometry import XmiBaseGeometry
from .xmi_point_3d import XmiPoint3D


class XmiArc3D(XmiBaseGeometry):
    start_point: XmiPoint3D = Field(..., description="Start point of the arc")
    end_point: XmiPoint3D = Field(..., description="End point of the arc")
    center_point: XmiPoint3D = Field(..., description="Center point of the arc")
    radius: Optional[float] = Field(None, description="Radius of the arc")

    @field_validator("start_point", "end_point", "center_point", mode="before")
    @classmethod
    def validate_points(cls, v, field):
        if not isinstance(v, XmiPoint3D):
            raise TypeError(f"{field.name} should be an XmiPoint3D")
        return v

    @model_validator(mode="before")
    @classmethod
    def ensure_required_fields(cls, values):
        missing = [attr for attr in ("start_point", "end_point", "center_point") if attr not in values]
        if missing:
            raise ValueError(f"{', '.join(missing)} are compulsory and must be provided.")
        return values

    @classmethod
    def from_dict(cls, obj: dict) -> Tuple[Optional["XmiArc3D"], List[Exception]]:
        error_logs: List[Exception] = []

        missing = [attr for attr in ("start_point", "end_point", "center_point") if attr not in obj]
        if missing:
            for attr in missing:
                error_logs.append(Exception(f"Missing attribute: {attr}"))
            return None, error_logs

        try:
            instance = cls(**obj)
        except Exception as e:
            error_logs.append(Exception(f"Error instantiating XmiArc3D: {e}"))
            instance = None

        return instance, error_logs


# Testing run python -m src.xmi.v2.models.geometries.xmi_arc_3d

if __name__ == "__main__":
    from .xmi_point_3d import XmiPoint3D

    p1 = XmiPoint3D(x=0, y=0, z=0)
    p2 = XmiPoint3D(x=1, y=0, z=0)
    p3 = XmiPoint3D(x=0.5, y=0.5, z=0)

    arc_data = {
        "start_point": p1,
        "end_point": p2,
        "center_point": p3,
        "radius": 0.5,
        "name": "Arc between P1 and P2"
    }

    arc, errors = XmiArc3D.from_dict(arc_data)
    if arc:
        print(arc.model_dump())
    else:
        for e in errors:
            print(f"Error: {e}")


from ..enums.xmi_shape_enums import XmiShapeEnum

# RECTANGULAR = (1, "Rectangular", 2, ("H", "B"))
# CIRCULAR = (2, "Circular", 1, ("D"))
# L_SHAPE = (3, "L Shape", 4, ("H", "B", "T", "t"))
# T_SHAPE = (4, "T Shape", 4, ("H", "B", "T", "t"))
# C_SHAPE = (5, "C Shape", 5, ("H", "B", "T1", "T2", "t"))
# I_SHAPE = (6, "I Shape", 5, ("D", "B", "T", "t", "r"))
# SQUARE_HOLLOW = (7, "Square Hollow", 2, ("D", "t"))
# RECTANGULAR_HOLLOW = (8, "Rectangular Hollow", 3, ("D", "B", "t"))
# OTHERS = (9, "Others", 0, ())


class XmiShape():

    def __init__(self, shape: XmiShapeEnum,
                 parameter_quantity: int,
                 parameter_names: tuple[str]) -> None:
        self.shape: XmiShapeEnum = shape
        self.parameter_quantity: int = parameter_quantity
        self.parameter_names: tuple[str] = parameter_names


class XmiShapeRectangle(XmiShape):
    def __init__(self) -> None:
        shape = XmiShapeEnum.RECTANGULAR
        parameter_quantity = 2
        parameter_names = ("H", "B")

        super().__init__(shape, parameter_quantity, parameter_names)


class XmiShapeCircle(XmiShape):
    def __init__(self) -> None:
        shape = XmiShapeEnum.CIRCULAR
        parameter_quantity = 1
        parameter_names = ("D")
        super().__init__(shape, parameter_quantity, parameter_names)


class XmiShapeL(XmiShape):
    def __init__(self) -> None:
        shape = XmiShapeEnum.L_SHAPE
        parameter_quantity = 4
        parameter_names = ("H", "B", "T", "t")
        super().__init__(shape, parameter_quantity, parameter_names)


class XmiShapeC(XmiShape):
    def __init__(self) -> None:
        shape = XmiShapeEnum.C_SHAPE
        parameter_quantity = 5
        parameter_names = ("H", "B", "T1", "T2", "t")
        super().__init__(shape, parameter_quantity, parameter_names)


class XmiShapeI(XmiShape):
    def __init__(self) -> None:
        shape = XmiShapeEnum.I_SHAPE
        parameter_quantity = 5
        parameter_names = ("D", "B", "T", "t", "r")
        super().__init__(shape, parameter_quantity, parameter_names)


class XmiShapeT(XmiShape):
    def __init__(self) -> None:
        shape = XmiShapeEnum.T_SHAPE
        parameter_quantity = 4
        parameter_names = ("H", "B", "T", "t")
        super().__init__(shape, parameter_quantity, parameter_names)


class XmiShapeSquareHollow(XmiShape):
    def __init__(self) -> None:
        shape = XmiShapeEnum.SQUARE_HOLLOW
        parameter_quantity = 2
        parameter_names = ("D", "t")
        super().__init__(shape, parameter_quantity, parameter_names)


class XmiShapeRectangularHollow(XmiShape):
    def __init__(self) -> None:
        shape = XmiShapeEnum.RECTANGULAR_HOLLOW
        parameter_quantity = 3
        parameter_names = ("D", "B", "t")
        super().__init__(shape, parameter_quantity, parameter_names)


class XmiShapeOthers(XmiShape):
    def __init__(self) -> None:
        shape = XmiShapeEnum.OTHERS
        parameter_quantity = 0
        parameter_names = tuple()
        super().__init__(shape, parameter_quantity, parameter_names)



# Optional, for forward declarations in Python 3.7+
from __future__ import annotations

import uuid


class XmiBaseEntity():

    __slots__ = ('_id',
                 '_name',
                 '_ifcguid',
                 '_description',
                 '_entity_type')

    def __init__(self,
                 id: str = None,
                 name: str = None,
                 ifcguid: str = None,
                 description: str = None,
                 entity_type: str = "XmiBaseEntity",
                 ** kwargs):

        id = id if id else kwargs.get('id', str(uuid.uuid4()))
        name = name if name else kwargs.get('name', id)
        description = description if description else kwargs.get(
            'description', None)
        ifcguid = ifcguid if ifcguid else kwargs.get('ifcguid', None)

        self.id = id  # Calls the setter
        self.name = name
        self.ifcguid = ifcguid
        self.description = description
        self.entity_type = entity_type

    @property
    def entity_type(self):
        return self._entity_type

    @entity_type.setter
    def entity_type(self, value):
        # if value is not isinstance(value, str):
        #     raise TypeError("'entity_type' should be an str")
        self._entity_type = value

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        if value is not None and not isinstance(value, str):
            raise TypeError("ID should be an str or None")
        self._id = value

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if value is not None and not isinstance(value, str):
            raise TypeError("Name should be a str or None")
        self._name = value

    @property
    def ifcguid(self):
        return self._ifcguid

    @ifcguid.setter
    def ifcguid(self, value):
        if value is not None and not isinstance(value, (str)):
            raise TypeError("IFCGUID should be a string")
        self._ifcguid = value

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        if value is not None and not isinstance(value, str):
            raise TypeError("Description should be a string")
        self._description = value

    def to_dict(self):
        return {slot.lstrip('_'): getattr(self, slot) for slot in self.__slots__}


class XmiBaseRelationship():
    __slots__ = ('_source', '_target', '_name', '_entity_type')

    def __init__(self,
                 source: XmiBaseEntity,
                 target: XmiBaseEntity,
                 name: str,
                 entity_type: str = "XmiRelBaseRelationship"
                 ) -> XmiBaseRelationship:
        """_summary_

        Parameters
        ----------
        source : XmiBaseEntity
            _description_
        target : XmiBaseEntity
            _description_
        name : str, optional
            _description_, by default None

        Returns
        -------
        XmiBaseRelationship
            _description_
        """

        if not isinstance(source, XmiBaseEntity):
            raise TypeError("'source' should be of type XmiBaseEntity")

        if not isinstance(target, XmiBaseEntity):
            raise TypeError("'target' should be of type XmiBaseEntity")

        if name is None:
            raise ValueError("'name' value should not be None")

        self.source = source
        self.target = target
        self.name = name
        self.entity_type = entity_type

    # Getter and setter for source
    @property
    def entity_type(self):
        return self._entity_type

    @entity_type.setter
    def entity_type(self, value):
        if not isinstance(value, str):
            raise ValueError("'entity_type' must be of type str")
        self._entity_type = value

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, source):
        if isinstance(source, XmiBaseEntity):
            self._source = source
        else:
            raise ValueError("Source must be of type XmiBaseEntity")

    # Getter and setter for target
    @property
    def target(self):
        return self._target

    @target.setter
    def target(self, target):
        if isinstance(target, XmiBaseEntity):
            self._target = target
        else:
            raise ValueError("Target must be of type XmiBaseEntity")

    # Getter and setter for name
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name



from pydantic import field_validator, model_validator
from ..bases.xmi_base_relationship import XmiBaseRelationship
from ..bases.xmi_base_entity import XmiBaseEntity
from ..entities.xmi_structural_cross_section import XmiStructuralCrossSection

class XmiHasStructuralCrossSection(XmiBaseRelationship):
    @field_validator("source", mode="before")
    @classmethod
    def validate_source(cls, v):
        if not isinstance(v, XmiBaseEntity):
            raise TypeError("Source must be of type XmiBaseEntity")
        return v

    @field_validator("target", mode="before")
    @classmethod
    def validate_target(cls, v):
        if not isinstance(v, XmiStructuralCrossSection):
            raise TypeError("Target must be of type XmiStructuralCrossSection")
        return v

    @model_validator(mode="before")
    @classmethod
    def set_defaults(cls, values):
        values.setdefault("name", "hasStructuralCrossSection")
        values.setdefault("entity_type", "XmiRelHasStructuralCrossSection")
        return values


# Testing run python -m src.xmi.v2.models.relationships.xmi_has_structural_cross_section
if __name__ == "__main__":
    from ..entities.xmi_structural_material import XmiStructuralMaterial
    from ..enums.xmi_structural_material_type_enum import XmiStructuralMaterialTypeEnum

    material = XmiStructuralMaterial(
        id="mat001",
        name="Steel S355",
        material_type=XmiStructuralMaterialTypeEnum.STEEL,
        grade=355.0,
        unit_weight=7850.0,
        e_modulus=200000.0,
        g_modulus=80000.0,
        poisson_ratio=0.3,
        thermal_coefficient=1.2e-5,
        description="High-strength steel"
    )

    cross_section = XmiStructuralCrossSection(
        id="cs001",
        name="I-Beam 300x150",
        shape="I Shape",
        parameters=(300, 150, 10, 6, 8),
        material=material,
        ix=1200,
        iy=500,
        rx=3.5,
        ry=1.2,
        ex=200000,
        ey=200000,
        zx=300,
        zy=150,
        j=80,
        area=4500,
        description="Test I-beam section",
        ifcguid="a-b-c-d"
    )

    source_entity = XmiBaseEntity(id="sm001", name="Member A", entity_type="XmiStructuralCurveMember")

    rel = XmiHasStructuralCrossSection(source=source_entity, target=cross_section)

    print("Created XmiHasStructuralCrossSection relationship:")
    print(rel.model_dump(by_alias=True))


from typing import List, Tuple, Optional
from pydantic import Field, field_validator, model_validator
from ..bases.xmi_base_geometry import XmiBaseGeometry
from .xmi_point_3d import XmiPoint3D

class XmiLine3D(XmiBaseGeometry):
    start_point: XmiPoint3D = Field(..., description="Start point of the line")
    end_point: XmiPoint3D = Field(..., description="End point of the line")

    @field_validator("start_point", "end_point", mode="before")
    @classmethod
    def check_is_point(cls, v, field):
        if not isinstance(v, XmiPoint3D):
            raise TypeError(f"{field.name} must be an XmiPoint3D")
        return v

    @model_validator(mode="before")
    @classmethod
    def ensure_required_fields(cls, values: dict):
        missing = [field for field in ("start_point", "end_point") if field not in values]
        if missing:
            raise ValueError(f"{', '.join(missing)} are compulsory and must be provided.")
        return values

    @classmethod
    def from_dict(cls, obj: dict) -> Tuple[Optional["XmiLine3D"], List[Exception]]:
        errors: List[Exception] = []

        missing = [attr for attr in ("start_point", "end_point") if attr not in obj]
        if missing:
            for attr in missing:
                errors.append(Exception(f"Missing required field: {attr}"))
            return None, errors

        try:
            instance = cls(**obj)
        except Exception as e:
            errors.append(e)
            instance = None

        return instance, errors


# Testing run python -m src.xmi.v2.models.geometries.xmi_line_3d

if __name__ == "__main__":
    point1 = XmiPoint3D(x=0.0, y=0.0, z=0.0, name="Start", id="p1")
    point2 = XmiPoint3D(x=10.0, y=5.0, z=2.0, name="End", id="p2")

    line_data = {
        "start_point": point1,
        "end_point": point2,
        "name": "TestLine",
        "id": "line123",
        "description": "A simple test line"
    }

    line_instance, errors = XmiLine3D.from_dict(line_data)

    if errors:
        print("Errors during instantiation:")
        for e in errors:
            print(" -", e)
    else:
        print("Line created successfully:")
        print(line_instance.model_dump(by_alias=True))


from pydantic import Field, field_validator
from typing import Tuple, List, Optional
from ..bases.xmi_base_geometry import XmiBaseGeometry

class XmiPoint3D(XmiBaseGeometry):
    x: float = Field(..., alias="X", description="X coordinate")
    y: float = Field(..., alias="Y", description="Y coordinate")
    z: float = Field(..., alias="Z", description="Z coordinate")

    @field_validator("x", "y", "z")
    @classmethod
    def check_coordinate(cls, v, field):
        if not isinstance(v, (int, float)):
            raise TypeError(f"{field.name} must be a number")
        return float(v)

    @classmethod
    def from_dict(cls, obj: dict) -> Tuple[Optional["XmiPoint3D"], List[Exception]]:
        error_logs: List[Exception] = []
        processed_data = obj.copy()

        required_attrs = ["X", "Y", "Z"]
        for attr in required_attrs:
            if attr not in processed_data:
                error_logs.append(Exception(f"Missing attribute: {attr}"))
                processed_data[attr] = None

        try:
            instance = cls.model_validate(processed_data)
        except Exception as e:
            error_logs.append(e)
            instance = None

        return instance, error_logs

    @classmethod
    def from_xmi_dict_obj(cls, xmi_dict_obj: dict) -> Tuple["XmiPoint3D", List[Exception]]:
        key_map = {
            "Name": "name",
            "X": "x",
            "Y": "y",
            "Z": "z",
            "Description": "description",
            "ID": "id",
            "IFCGUID": "ifcguid",
        }

        processed = {key_map.get(k, k): v for k, v in xmi_dict_obj.items()}
        return cls.from_dict(processed)
    

# Testing run python -m src.xmi.v2.models.geometries.xmi_point_3d
    
if __name__ == "__main__":
    print("=== Running Quick Tests for XmiPoint3D ===")

    # Test 1: Valid input
    valid_data = {"x": 1.0, "y": 2.0, "z": 3.0}
    point, errors = XmiPoint3D.from_dict(valid_data)
    print("\n[Valid Input]")
    print("Instance:", point)
    print("Errors:", errors)

    # Test 2: Missing 'z'
    missing_z = {"x": 1.0, "y": 2.0}
    point, errors = XmiPoint3D.from_dict(missing_z)
    print("\n[Missing 'z']")
    print("Instance:", point)
    print("Errors:", [str(e) for e in errors])

    # Test 3: Invalid type for 'x'
    invalid_x = {"x": "a", "y": 2.0, "z": 3.0}
    point, errors = XmiPoint3D.from_dict(invalid_x)
    print("\n[Invalid 'x' Type]")
    print("Instance:", point)
    print("Errors:", [str(e) for e in errors])

    # Test 4: Valid XMI-formatted dict
    xmi_data = {
        "ID": "001",
        "Name": "PointA",
        "X": 10,
        "Y": 20,
        "Z": 30,
        "Description": "Example point"
    }
    point, errors = XmiPoint3D.from_xmi_dict_obj(xmi_data)
    print("\n[XMI-formatted Input]")
    print("Instance:", point.model_dump() if point else None)
    print("Errors:", [str(e) for e in errors])


from pydantic import Field, model_validator
from typing import Optional
from .xmi_base_entity import XmiBaseEntity

class XmiBaseGeometry(XmiBaseEntity):
    entity_type: Optional[str] = Field(None, alias="EntityType")

    @model_validator(mode="before")
    @classmethod
    def set_entity_type(cls, values):
        values["EntityType"] = values.get("EntityType") or values.get("entity_type") or cls.__name__

        return values

    class Config:
        populate_by_name = True


# Testing run python -m src.xmi.v2.models.bases.xmi_base_geometry

geometry = XmiBaseGeometry(
    ID="001",
    Name=None,
    IFCGUID="abc-123",
    Description="Test geometry"
)

print(geometry.model_dump(by_alias=True, exclude_none=True))


# Optional, for forward declarations in Python 3.7+
from __future__ import annotations

from ..xmi_base import XmiBaseRelationship, XmiBaseEntity
from ..constants import *


class XmiHasPoint3D(XmiBaseRelationship):
    __slots__ = XmiBaseRelationship.__slots__

    _attributes_needed = [slot[1:] if slot.startswith(
        '_') else slot for slot in __slots__]

    def __init__(self, source: XmiBaseEntity, target: XmiBaseEntity, name='hasPoint3D', **kwargs):
        name = 'hasPoint3D'
        entity_type = "XmiRelHasPoint3D"

        super().__init__(source, target, name, entity_type=entity_type)

        for key, value in kwargs.items():
            if key in self.__slots__:
                # Use the property setter for type checking
                setattr(self, key, value)



# Optional, for forward declarations in Python 3.7+
from __future__ import annotations

from .xmi_base import XmiBaseEntity, XmiBaseRelationship


class ErrorLog():
    def __init__(self, entity_type: str, index: int, message: str, obj: str = None):
        self.entity_type: str = entity_type
        self.index: int = index
        self.message: str = message
        self.obj: str = str(obj)


class XmiModel():
    def __init__(self,
                 name: str = None,
                 xmi_version: str = None,
                 application_name: str = None,
                 application_version: str = None
                 ):

        self.entities: list[XmiBaseEntity] = []
        self.relationships: list[XmiBaseRelationship] = []
        self.histories = []
        self.errors: list[ErrorLog] = []
        self.name = name
        self.xmi_version = xmi_version
        self.application_name = application_name
        self.application_version = application_version

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if value is not None and not isinstance(value, str):
            raise TypeError("'name' should be an str or None")
        self._name = value

    @property
    def xmi_version(self):
        return self._xmi_version

    @xmi_version.setter
    def xmi_version(self, value):
        if value is not None and not isinstance(value, (str)):
            raise TypeError("'xmi_version' should be a string")
        self._xmi_version = value

    @property
    def application_name(self):
        return self._application_name

    @application_name.setter
    def application_name(self, value):
        if value is not None and not isinstance(value, str):
            raise TypeError("'application_name' should be a string")
        self._application_name = value

    @property
    def application_version(self):
        return self._application_version

    @application_version.setter
    def application_version(self, value):
        if value is not None and not isinstance(value, str):
            raise TypeError("'application_version' should be a string")
        self._application_version = value

    def create_relationship(self,
                            relationship_class: XmiBaseRelationship,
                            source: XmiBaseEntity,
                            target: XmiBaseEntity,
                            name: str = None,
                            **kwargs) -> XmiBaseRelationship:

        if relationship_class == XmiBaseRelationship:
            relationship = relationship_class(source, target, name)
        else:
            relationship = relationship_class(
                source, target, **kwargs)
        self.relationships.append(relationship)
        return relationship

    def create_entity(self, entity_class: XmiBaseEntity, **kwargs) -> XmiBaseEntity:
        entity = entity_class(**kwargs)
        self.entities.append(entity)
        return entity

    def find_relationships_by_target(self, target: XmiBaseEntity) -> list[XmiBaseRelationship]:
        return [rel for rel in self.relationships if rel.target == target]

    def find_relationships_by_source(self, source: XmiBaseEntity) -> list[XmiBaseRelationship]:
        return [rel for rel in self.relationships if rel.source == source]



from .xmi_enums import XmiEnum
from enum import unique


@unique
class XmiStructuralMaterialTypeEnum(XmiEnum):
    CONCRETE = "Concrete"
    STEEL = "Steel"
    TIMBER = "Timber"
    ALUMINIUM = "Aluminium"
    COMPOSITE = "Composite"
    MASONRY = "Masonry"
    OTHERS = "Others"
    REBAR = "Rebar"  # to be removed
    TENDON = "Tendon"  # to be removed



from pydantic import Field, field_validator, model_validator
from typing import Optional, Tuple, List
from ..bases.xmi_base_entity import XmiBaseEntity
from ..enums.xmi_structural_material_type_enum import XmiStructuralMaterialTypeEnum


class XmiStructuralMaterial(XmiBaseEntity):
    material_type: XmiStructuralMaterialTypeEnum = Field(..., alias="MaterialType")
    grade: Optional[float] = Field(None, alias="Grade")
    unit_weight: Optional[float] = Field(None, alias="UnitWeight")
    e_modulus: Optional[float] = Field(None, alias="EModulus")
    g_modulus: Optional[float] = Field(None, alias="GModulus")
    poisson_ratio: Optional[float] = Field(None, alias="PoissonRatio")
    thermal_coefficient: Optional[float] = Field(None, alias="ThermalCoefficient")

    class Config:
        populate_by_name = True

    @field_validator("material_type")
    @classmethod
    def validate_material_type(cls, v):
        if not isinstance(v, XmiStructuralMaterialTypeEnum):
            raise TypeError("material_type must be a valid XmiStructuralMaterialTypeEnum")
        return v

    @field_validator("grade", "unit_weight", "e_modulus", "g_modulus", "poisson_ratio", "thermal_coefficient")
    @classmethod
    def validate_floats(cls, v):
        if v is not None and not isinstance(v, (float, int)):
            raise TypeError("Value must be float, int, or None")
        return v

    @model_validator(mode="before")
    @classmethod
    def set_entity_type(cls, values):
        values.setdefault("entity_type", "XmiStructuralMaterial")
        return values

    '''@classmethod
    def from_dict(cls, obj: dict) -> Tuple[Optional["XmiStructuralMaterial"], List[Exception]]:
        error_logs: List[Exception] = []
        required = ["material_type"]
        processed = obj.copy()

        for attr in required:
            if attr not in processed:
                error_logs.append(Exception(f"Missing attribute: {attr}"))
                processed[attr] = None

        try:
            processed["material_type"] = XmiStructuralMaterialTypeEnum.from_attribute_get_enum(
                processed.get("material_type")
            )
        except Exception as e:
            error_logs.append(e)

        try:
            instance = cls.model_validate(processed)
        except Exception as e:
            error_logs.append(Exception(f"Error instantiating XmiStructuralMaterial: {e}"))
            instance = None

        return instance, error_logs

    @classmethod
    def from_xmi_dict_obj(cls, xmi_dict_obj: dict) -> Tuple[Optional["XmiStructuralMaterial"], List[Exception]]:
        key_map = {
            "Name": "name",
            "Type": "material_type",
            "Grade": "grade",
            "UnitWeight": "unit_weight",
            "EModulus": "e_modulus",
            "GModulus": "g_modulus",
            "PoissonRatio": "poisson_ratio",
            "Description": "description",
            "ID": "id",
            "IFCGUID": "ifcguid",
            "ThermalCoefficient": "thermal_coefficient",
        }

        processed = {key_map.get(k, k): v for k, v in xmi_dict_obj.items()}
        return cls.from_dict(processed)'''


# Testing run python -m src.xmi.v2.models.entities.xmi_structural_material
if __name__ == "__main__":
    def test_structural_material():
        test_data = {
            "material_type": "Steel",
            "grade": 450,
            "unit_weight": 7850,
            "e_modulus": 200000,
            "g_modulus": 80000,
            "poisson_ratio": 0.3,
            "thermal_coefficient": 1.2e-5,
            "name": "S450",
            "description": "High-strength steel",
            "id": "mat001"
        }

        instance, errors = XmiStructuralMaterial.from_dict(test_data)

        if errors:
            print("Errors encountered:")
            for e in errors:
                print("-", e)
        else:
            print("Created XmiStructuralMaterial:")
            print(instance.model_dump_json(by_alias=True))

    test_structural_material()


# Optional, for forward declarations in Python 3.7+
from __future__ import annotations

from ..xmi_base import XmiBaseEntity
from .xmi_base_geometry import XmiBaseGeometry


class XmiPoint3D(XmiBaseGeometry):

    __slots__ = XmiBaseEntity.__slots__ + ('_x', '_y', '_z')

    _attributes_needed = [slot[1:] if slot.startswith(
        '_') else slot for slot in __slots__ if slot != "_entity_type"]

    def __init__(self,
                 x: float,
                 y: float,
                 z: float,
                 id: str = None,
                 name: str = None,
                 description: str = None,
                 ifcguid: str = None,
                 **kwargs):
        entity_type = "XmiPoint3D"

        # Ensure material_type is provided
        if x is None:
            raise ValueError(
                "The 'x' parameter is compulsory and must be provided.")
        # Ensure material_type is provided
        if y is None:
            raise ValueError(
                "The 'y' parameter is compulsory and must be provided.")
        # Ensure material_type is provided
        if z is None:
            raise ValueError(
                "The 'z' parameter is compulsory and must be provided.")

        # Initialize parent class
        super().__init__(id=id,
                         name=name,
                         ifcguid=ifcguid,
                         description=description,
                         entity_type=entity_type
                         )

        # Initialize attributes
        self.set_attributes(x, y, z, **kwargs)

    def set_attributes(self, x, y, z, **kwargs):
        attributes = [
            ('x', x),
            ('y', y),
            ('z', z),
        ]

        for attr_name, attr_value in attributes:
            value = kwargs.get(attr_name, attr_value)
            try:
                setattr(self, attr_name, value)
            except AttributeError as e:
                print(
                    f"Caught an AttributeError while setting {attr_name}: {e}")
                setattr(self, attr_name, None)

    @property
    def x(self):
        return self._x

    @x.setter
    def x(self, value):
        if not isinstance(value, (int, float)):
            raise TypeError("X should be an int or float")
        self._x = value

    @property
    def y(self):
        return self._y

    @y.setter
    def y(self, value):
        if not isinstance(value, (int, float)):
            raise TypeError("Y should be an int or float")
        self._y = value

    @property
    def z(self):
        return self._z

    @z.setter
    def z(self, value):
        if not isinstance(value, (int, float)):
            raise TypeError("Z should be an int or float")
        self._z = value

    @property
    def storey(self):
        return self._storey

    @storey.setter
    def storey(self, value):
        if not isinstance(value, str):
            raise TypeError("Storey should be an str")
        self._storey = value

    @classmethod
    def from_dict(cls, obj: dict) -> XmiPoint3D:
        error_logs = []
        processed_data = obj.copy()

        for attr in cls._attributes_needed:
            if attr not in obj:
                error_logs.append(Exception(f"Missing attribute: {attr}"))
                processed_data[attr] = None

        x_found = processed_data['x']
        y_found = processed_data['y']
        z_found = processed_data['z']

        # remove compulsory keys for proper class instantiation
        del processed_data['x']
        del processed_data['y']
        del processed_data['z']
        try:
            instance = cls(
                x=x_found,
                y=y_found,
                z=z_found,
                **processed_data)
        except Exception as e:
            error_logs.append(
                Exception(f"Error instantiating XmiPoint3D: {obj}"))

        return instance, error_logs

    @classmethod
    def from_xmi_dict_obj(cls, xmi_dict_obj: dict):
        # currently xmi_file doesnt support xmi id for reference, will assume structural_point_connection's id and name

        # Define a mapping from snake_case keys to custom keys
        KEY_MAPPING = {
            "Name": "name",
            "X": "x",
            "Y": "y",
            "Z": "z",
            "Description": "description",
            "ID": "id",
            "IFCGUID": "ifcguid",
        }
        instance = None
        error_logs = []
        processed_data = {KEY_MAPPING.get(
            key, key): value for key, value in xmi_dict_obj.items() if key in KEY_MAPPING}

        instance, error_logs_found = cls.from_dict(
            processed_data)

        error_logs.extend(error_logs_found)

        return instance, error_logs



from enum import unique
from ..bases.xmi_base_enum import XmiBaseEnum

@unique
class XmiStructuralSurfaceMemberSystemPlaneEnum(XmiBaseEnum):
    BOTTOM = "Bottom"
    TOP = "Top"
    MIDDLE = "Middle"
    LEFT = "Left"
    RIGHT = "Right"


# Testing run python -m src.xmi.v2.models.enums.xmi_structural_surface_member_system_plane_enum

print(XmiStructuralSurfaceMemberSystemPlaneEnum.RIGHT.value)
print(XmiStructuralSurfaceMemberSystemPlaneEnum.TOP.name)


from enum import Enum, unique
from typing import Optional

@unique
class XmiBaseEnum(str, Enum): 
    @classmethod
    def _missing_(cls, value):
        if not isinstance(value, str):
            return None
        for member in cls:
            if member.value.lower() == value.lower():
                return member
        return None
    
    @classmethod
    def from_name_get_enum(cls, name_str: str) -> Optional["XmiBaseEnum"]:
        name_str = name_str.upper()
        try:
            return cls[name_str]
        except KeyError:
            return None

    @classmethod
    def from_attribute_get_enum(cls, attribute_str: str) -> Optional["XmiBaseEnum"]:
        for member in cls:
            if member.value == attribute_str:
                return member
        return None


from enum import unique
from ..bases.xmi_base_enum import XmiBaseEnum

@unique
class XmiStructuralCurveMemberTypeEnum(XmiBaseEnum):
    BEAM = "Beam"
    COLUMN = "Column"
    BRACING = "Bracing"
    OTHER = "Other"


# Testing run python -m src.xmi.v2.models.enums.xmi_structural_curve_member_type_enum

print(XmiStructuralCurveMemberTypeEnum.BRACING.value)
print(XmiStructuralCurveMemberTypeEnum.COLUMN.name)


from abc import ABC
from pydantic import BaseModel, Field, model_validator
from typing import Optional
import uuid

class XmiBaseEntity(BaseModel, ABC):
    id: str = Field(default_factory=lambda: str(uuid.uuid4()), alias="ID")
    name: Optional[str] = Field(None, alias="Name")
    ifcguid: Optional[str] = Field(None, alias="IFCGUID")
    native_id: Optional[str] = Field(None, alias="NativeId")
    description: Optional[str] = Field(None, alias="Description")
    entity_type: Optional[str] = Field(None, alias="EntityType")

    @model_validator(mode="before")
    @classmethod
    def fill_defaults(cls, values: dict):
        id_ = values.get("ID") or values.get("id") or str(uuid.uuid4())

        if "ID" not in values and "id" not in values:
            values["ID"] = id_

        if "Name" not in values and "name" not in values:
            values["Name"] = id_

        if "EntityType" not in values and "entity_type" not in values:
            values["EntityType"] = cls.__name__

        return values

    class Config:
        populate_by_name = True


# Testing run python -m src.xmi.v2.models.bases.xmi_base_entity 

class XmiStructuralMaterial(XmiBaseEntity):
    material_type: str
    unit_weight: Optional[float] = None

    def get_info(self) -> str:
        return f"id: {self.id}, {self.name} ({self.material_type}), weight: {self.unit_weight}"

    def create(self):
        print(f"Creating structural material: {self.name}")

entity = XmiStructuralMaterial(
    name="123",
    ifcguid="ABC123",
    description="Steel beam",
    material_type="Steel",
    unit_weight=78.5
)

print(entity.get_info())
entity.create()


# Optional, for forward declarations in Python 3.7+
from __future__ import annotations

from ..enums.xmi_enums import XmiSegmentTypeEnum

from ..entities.xmi_structural_point_connection import XmiStructuralPointConnection

from ..geometries.xmi_base_geometry import XmiBaseGeometry

from ..xmi_base import XmiBaseEntity


class XmiSegment(XmiBaseEntity):
    __slots__ = XmiBaseEntity.__slots__ + ('_geometry',
                                           '_position',
                                           '_begin_node',
                                           '_end_node',
                                           '_segment_type'
                                           )

    _attributes_needed = [slot[1:] if slot.startswith(
        '_') else slot for slot in __slots__ if slot != "_entity_type"]

    def __init__(self,
                 geometry: XmiBaseGeometry,
                 position: int,
                 begin_node: XmiStructuralPointConnection,
                 end_node: XmiStructuralPointConnection,
                 segment_type: XmiSegmentTypeEnum,
                 id: str = None,
                 name: str = None,
                 description: str = None,
                 ifcguid: str = None,
                 **kwargs):
        entity_type = "XmiSegment"

        # Ensure geometry is provided
        if geometry is None:
            raise ValueError(
                "The 'geometry' parameter is compulsory and must be provided.")

        # Ensure position is provided
        if position is None:
            raise ValueError(
                "The 'position' parameter is compulsory and must be provided.")

        # Ensure begin_node is provided
        if begin_node is None:
            raise ValueError(
                "The 'start_node' parameter is compulsory and must be provided.")

        # Ensure end_node is provided
        if end_node is None:
            raise ValueError(
                "The 'end_node' parameter is compulsory and must be provided.")

        # Ensure segment_type is provided
        if segment_type is None:
            raise ValueError(
                "The 'segment_type' parameter is compulsory and must be provided.")

        # Initialize parent class
        super().__init__(id=id,
                         name=name,
                         ifcguid=ifcguid,
                         description=description,
                         entity_type=entity_type
                         )

        # Initialize attributes
        self.set_attributes(
            geometry,
            position,
            begin_node,
            end_node,
            segment_type,
            **kwargs)

    def set_attributes(self,
                       geometry: XmiBaseGeometry,
                       position: int,
                       begin_node: XmiStructuralPointConnection,
                       end_node: XmiStructuralPointConnection,
                       segment_type: XmiSegmentTypeEnum,
                       **kwargs):
        attributes = [
            ('geometry', geometry),
            ('position', position),
            ('begin_node', begin_node),
            ('end_node', end_node),
            ('segment_type', segment_type)
        ]

        for attr_name, attr_value in attributes:
            value = kwargs.get(attr_name, attr_value)
            try:
                setattr(self, attr_name, value)
            except AttributeError as e:
                print(
                    f"Caught an AttributeError while setting {attr_name}: {e}")
                setattr(self, attr_name, None)

    @property
    def geometry(self):
        return self._geometry

    @geometry.setter
    def geometry(self, value):
        if not isinstance(value, XmiBaseGeometry):
            raise TypeError(
                "geometry should be an XmiBaseGeometry")
        self._geometry = value

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, value):
        if not isinstance(value, int):
            raise TypeError(
                "position should be of type int")
        self._position = value

    @property
    def begin_node(self):
        return self._begin_node

    @begin_node.setter
    def begin_node(self, value):
        if not isinstance(value, XmiStructuralPointConnection):
            raise TypeError(
                "begin_node should be of type XmiStructuralPointConnection")
        self._begin_node = value

    @property
    def end_node(self):
        return self._end_node

    @end_node.setter
    def end_node(self, value):
        if not isinstance(value, XmiStructuralPointConnection):
            raise TypeError(
                "end_node should be of type XmiStructuralPointConnection")
        self._end_node = value

    @property
    def segment_type(self):
        return self._segment_type

    @segment_type.setter
    def segment_type(self, value):
        if not isinstance(value, XmiSegmentTypeEnum):
            raise TypeError(
                "segment_type should be of type XmiSegmentTypeEnum")
        self._segment_type = value



from pydantic import field_validator, model_validator
from ..bases.xmi_base_relationship import XmiBaseRelationship
from ..bases.xmi_base_entity import XmiBaseEntity
from ..entities.xmi_segment import XmiSegment


class XmiHasSegment(XmiBaseRelationship):
    @field_validator("source", mode="before")
    @classmethod
    def validate_source(cls, v):
        if not isinstance(v, XmiBaseEntity):
            raise TypeError("'source' must be an instance of XmiBaseEntity")
        return v

    @field_validator("target", mode="before")
    @classmethod
    def validate_target(cls, v):
        if not isinstance(v, XmiSegment):
            raise TypeError("'target' must be an instance of XmiBaseEntity")
        return v

    @model_validator(mode="before")
    @classmethod
    def set_defaults(cls, values):
        values.setdefault("name", "hasSegment")
        values.setdefault("entity_type", "XmiRelHasSegment")
        return values


# Testing run python -m src.xmi.v2.models.relationships.xmi_has_segment
if __name__ == "__main__":
    from ..enums.xmi_segment_type_enum import XmiSegmentTypeEnum
    from ..geometries.xmi_point_3d import XmiPoint3D
    from ..bases.xmi_base_geometry import XmiBaseGeometry
    from ..entities.xmi_structural_point_connection import XmiStructuralPointConnection
    from ..entities.xmi_structural_storey import XmiStructuralStorey

    geometry = XmiBaseGeometry(id="g1", name="Geometry1")
    storey = XmiStructuralStorey(id="s1", name="Storey1", storey_elevation=0.0)

    point = XmiPoint3D(x=0.0, y=0.0, z=0.0)
    begin_node = XmiStructuralPointConnection(id="n1", name="Begin", point=point, storey=storey)
    end_node = XmiStructuralPointConnection(id="n2", name="End", point=point, storey=storey)

    target = XmiSegment(
        id="E2",
        name="Segment1",
        geometry=geometry,
        position=1,
        begin_node=begin_node,
        end_node=end_node,
        segment_type=XmiSegmentTypeEnum.LINE
    )

    source = XmiBaseEntity(id="E1", name="MainElement", entity_type="StructuralElement")

    rel = XmiHasSegment(source=source, target=target)

    print("Created XmiHasSegment relationship:")
    print(rel.model_dump(by_alias=True))


from pydantic import Field
from ..bases.xmi_base_entity import XmiBaseEntity

class XmiStructuralStorey(XmiBaseEntity):
    storey_elevation: float = Field(..., alias="StoreyElevation")

	

# Testing run python -m src.xmi.v2.models.entities.xmi_structural_storey
if __name__ == "__main__":
	def test_xmi_structural_storey():
		storey = XmiStructuralStorey(
            id="storey001",
            name="Ground Floor",
            storey_elevation=0.0
        )
		print("Created XmiStructuralStorey:")
		print(storey.model_dump(by_alias=True))
		
	test_xmi_structural_storey()



from pydantic import Field, field_validator, model_validator
from typing import Optional, Tuple, Union, List, Dict, Any
from ..bases.xmi_base_entity import XmiBaseEntity
from ..entities.xmi_structural_material import XmiStructuralMaterial
from ..enums.xmi_shape_enum import XmiShapeEnum
from ...utils.xmi_utilities import is_empty_or_whitespace
from ...utils.xmi_errors import XmiInconsistentDataTypeError, XmiMissingRequiredAttributeError

class XmiStructuralCrossSection(XmiBaseEntity):
    material: XmiStructuralMaterial = Field(..., alias="Material")
    shape: XmiShapeEnum = Field(..., alias="Shape")
    parameters: Tuple[Union[float, int], ...] = Field(..., alias="Parameters")
    area: Optional[float] = Field(None, alias="Area")
    second_moment_of_area_x_axis: Optional[float] = Field(None, alias="SecondMomentOfAreaXAxis")
    second_moment_of_area_y_axis: Optional[float] = Field(None, alias="SecondMomentOfAreaYAxis")
    radius_of_gyration_x_axis: Optional[float] = Field(None, alias="RadiusOfGyrationXAxis")
    radius_of_gyration_y_axis: Optional[float] = Field(None, alias="RadiusOfGyrationYAxis")
    elastic_modulus_x_axis: Optional[float] = Field(None, alias="ElasticModulusXAxis")
    elastic_modulus_y_axis: Optional[float] = Field(None, alias="ElasticModulusYAxis")
    plastic_modulus_x_axis: Optional[float] = Field(None, alias="PlasticModulusXAxis")
    plastic_modulus_y_axis: Optional[float] = Field(None, alias="PlasticModulusYAxis")
    torsional_constant: Optional[float] = Field(None, alias="TorsionalConstant")
    
    class Config:
        populate_by_name = True

    @field_validator("parameters")
    @classmethod
    def validate_parameters(cls, v):
        if not isinstance(v, (list, tuple)):
            raise TypeError("parameters must be a list or tuple")
        for item in v:
            if not isinstance(item, (int, float)):
                raise TypeError("Each parameter must be an int or float")
            if item < 0:
                raise ValueError("Value cannot be smaller than 0")
        return tuple(v)

    @field_validator("material")
    @classmethod
    def validate_material(cls, v):
        if not isinstance(v, XmiStructuralMaterial):
            raise TypeError("material must be of type XmiStructuralMaterial")
        return v

    @field_validator("shape")
    @classmethod
    def validate_shape(cls, v):
        if not isinstance(v, XmiShapeEnum):
            raise TypeError("shape must be of type XmiShapeEnum")
        return v
    
    @field_validator(
            "area", "torsional_constant", "second_moment_of_area_x_axis", "second_moment_of_area_y_axis", 
            "radius_of_gyration_x_axis", "radius_of_gyration_y_axis", "elastic_modulus_x_axis",
            "elastic_modulus_y_axis", "plastic_modulus_x_axis", "plastic_modulus_y_axis"
        )
    @classmethod
    def validate_positive_or_none(cls, v, info):
        if v is not None and v < 0:
            raise ValueError(f"{info.field_name} must be non-negative")
        return v

    @model_validator(mode="before")
    @classmethod
    def set_entity_type(cls, values):
        values.setdefault("entity_type", "XmiStructuralCrossSection")
        return values
    
    '''@classmethod
    def convert_parameter_string_to_tuple(cls, parameter_str: str) -> Tuple[float, ...]:
        parameter_list: List[str] = parameter_str.split(';')
        for param in parameter_list:
            if is_empty_or_whitespace(param):
                raise XmiInconsistentDataTypeError(
                    f"The parameter [{param}] should not be empty or whitespace")
            try:
                float(param)
            except ValueError:
                raise XmiInconsistentDataTypeError(
                    f"The parameter [{param}] within the XmiStructuralCrossSection 'parameters' attribute should be convertible to float")
        return tuple(float(param) for param in parameter_list)

    @classmethod
    def from_dict(cls, obj: Dict[str, Any]) -> Tuple[Optional["XmiStructuralCrossSection"], List[Exception]]:
        errors = []
        processed = obj.copy()

        if "shape" in processed:
            shape = XmiShapeEnum.from_attribute_get_enum(processed["shape"])
            if shape is None:
                errors.append(Exception(f"Invalid shape enum value: {processed['shape']}"))
                return None, errors
            processed["shape"] = shape
        else:
            errors.append(XmiMissingRequiredAttributeError("Missing 'shape'"))
            return None, errors

        if "parameters" in processed:
            try:
                processed["parameters"] = cls.convert_parameter_string_to_tuple(processed["parameters"])
            except Exception as e:
                errors.append(e)
                return None, errors
        else:
            errors.append(XmiMissingRequiredAttributeError("Missing 'parameters'"))
            return None, errors

        try:
            instance = cls(**processed)
        except Exception as e:
            errors.append(e)
            instance = None

        return instance, errors

    
    @classmethod
    def from_xmi_dict_obj(
        cls,
        xmi_dict_obj: Dict[str, Any],
        material: Optional[XmiStructuralMaterial] = None
    ) -> Tuple[Optional["XmiStructuralCrossSection"], List[Exception]]:
        key_map = {
            "Name": "name",
            "Material": "material",
            "Parameters": "parameters",
            "Shape": "shape",
            "Ix": "second_moment_of_area_x_axis",
            "Iy": "second_moment_of_area_y_axis",
            "rx": "radius_of_gyration_x_axis",
            "ry": "radius_of_gyration_y_axis",
            "Ex": "elastic_modulus_x_axis",
            "Ey": "elastic_modulus_y_axis",
            "Zx": "plastic_modulus_x_axis",
            "Zy": "plastic_modulus_y_axis",
            "J": "torsional_constant",
            "Area": "area",
            "Description": "description",
            "ID": "id",
            "IFCGUID": "ifcguid",
        }

        instance: XmiStructuralCrossSection | None = None
        error_logs: list[Exception] = []
        processed = {key_map.get(k, k): v for k, v in xmi_dict_obj.items()}

        if material is not None:
            processed["material"] = material

        instance, error_logs_found = cls.from_dict(processed)
        error_logs.extend(error_logs_found)

        return instance, error_logs'''
    

# Testing run python -m src.xmi.v2.models.entities.xmi_structural_cross_section
if __name__ == "__main__":
    from ..enums.xmi_structural_material_type_enum import XmiStructuralMaterialTypeEnum

    material = XmiStructuralMaterial(
        id="mat001",
        name="S450",
        material_type=XmiStructuralMaterialTypeEnum.STEEL,
        grade=450.0,
        unit_weight=7850.0,
        e_modulus=200000.0,
        g_modulus=80000.0,
        poisson_ratio=0.3,
        thermal_coefficient=1.2e-5,
        description="High-strength steel"
    )

    cross_section = XmiStructuralCrossSection(
        id="xs001",
        name="I-Beam 300x150",
        description="Test I-beam section",
        ifcguid="a-b-c-d",
        shape="I Shape",
        parameters=(300.0, 150.0, 10.0, 6.0, 8.0),
        material=material,
        ix=1200,
        iy=500,
        rx=3.5,
        ry=1.2,
        ex=200000,
        ey=200000,
        zx=300,
        zy=150,
        j=80,
        area=4500
    )

    print("Created XmiStructuralCrossSection:")
    print(cross_section.model_dump(by_alias=True))